<p class="logo">
  <a href="{{ pathto(master_doc) }}">
    <img class="logo" src="{{ pathto('_static/pipenv.png', 1) }}" title="https://kennethreitz.org/tattoos"/>
  </a>
</p>

<p>
  <iframe src="https://ghbtns.com/github-btn.html?user=pypa&repo=pipenv&type=watch&count=true&size=large"
    allowtransparency="true" frameborder="0" scrolling="0" width="200px" height="35px"></iframe>
</p>

<p>
  <strong>Pipenv</strong> is a production-ready tool that aims to bring the best of all packaging worlds to the Python world. It harnesses Pipfile, pip, and virtualenv into one single command.
  <p>It features very pretty terminal colors.</p>
</p>

<h3>Stay Informed</h3>
<p>Receive updates on new releases and upcoming projects.</p>

<p><iframe src="http://ghbtns.com/github-btn.html?user=kennethreitz&type=follow&count=true"
  allowtransparency="true" frameborder="0" scrolling="0" width="200" height="20"></iframe></p>

<p><a href="https://twitter.com/kennethreitz" class="twitter-follow-button" data-show-count="false">Follow @kennethreitz</a> <script>!function(d,s,id){var js,fjs=d.getElementsByTagName(s)[0],p=/^http:/.test(d.location)?'http':'https';if(!d.getElementById(id)){js=d.createElement(s);js.id=id;js.src=p+'://platform.twitter.com/widgets.js';fjs.parentNode.insertBefore(js,fjs);}}(document, 'script', 'twitter-wjs');</script></p>
<p><a href="https://saythanks.io/to/kennethreitz">Say Thanks!</a></p>
<p><a href="http://tinyletter.com/kennethreitz">Join Mailing List</a>.</p>

<h3>Other Projects</h3>

<p>More <a href="http://kennethreitz.org/">Kenneth Reitz</a> projects:</p>
<ul>
    <li><a href="http://pep8.org/">pep8.org</a></li>
    <li><a href="http://httpbin.org/">httpbin.org</a></li>
    <li><a href="https://howtopython.org">HowToPython.org</a></li>
    <li><a href="http://python-guide.org">The Python Guide</a></li>
    <li><a href="http://python-requests.org/">Requests: HTTP for Humans</a></li>
    <li><a href="https://github.com/kennethreitz/records">Records: SQL for Humans</a></li>
    <li><a href="http://www.git-legit.org">Legit: Git for Humans</a></li>
    <li><a href="http://docs.python-tablib.org/en/latest/">Tablib: Tabular Datasets</a></li>
    <li><a href="http://markdownplease.com">Markdown, Please!</a></li>
</ul>


<h3>Useful Links</h3>
<ul>
  <li><a href="http://docs.python-requests.org/en/latest/community/recommended/">Recommended Packages and Extensions</a></li>

  <p></p>

  <li><a href="http://github.com/pypa/pipenv">Pipenv @ GitHub</a></li>
  <li><a href="http://pypi.python.org/pypi/pipenv">Pipenv @ PyPI</a></li>
  <li><a href="https://launchpad.net/~pypa/+archive/ubuntu/ppa">Pipenv PPA (PyPA)</a></li>
  <li><a href="http://github.com/kennethreitz/pipenv/issues">Issue Tracker</a></li>
  <hr>
  <li><a href="http://pipenv-ja.readthedocs.io/ja/translate-ja/">日本語</a></li>
</ul>

